#define STRICT
#define _WIN32_WINDOWS 0x0500

#include <Windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <math.h>
#include <tchar.h>
#include <D3D9.h>
#include <D3DX9.h>
#include "DXUtil.h"
#include "D3DEnumeration.h"
#include "D3DSettings.h"
#include "D3DApp.h"
#include "D3DFile.h"
#include "D3DFont.h"
#include "D3DUtil.h"
#include "Frogger.h"

const int cWidth = 160;
const int cHeight = 120;

//-----------------------------------------------------------------------------
// Name: class CMyD3DApplication
// Desc: Main class to run this application. Most functionality is inherited
//       from the CD3DApplication base class.
//-----------------------------------------------------------------------------
class CMyD3DApplication : public CD3DApplication
{
	// Scene description
	CFrogger*		m_Frogger;

public:
    CMyD3DApplication();

    HRESULT OneTimeSceneInit();
    HRESULT InitDeviceObjects();
    HRESULT RestoreDeviceObjects();
    HRESULT InvalidateDeviceObjects();
    HRESULT DeleteDeviceObjects();
    HRESULT Render();
    HRESULT FrameMove();
    HRESULT FinalCleanup();
    HRESULT ConfirmDevice(D3DCAPS9* pCaps, DWORD dwBehavior, D3DFORMAT adapterFormat, D3DFORMAT backBufferFormat);

    LRESULT MsgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
};

//-----------------------------------------------------------------------------
// Name: CMyD3DApplication()
// Desc: Constructor
//-----------------------------------------------------------------------------
CMyD3DApplication::CMyD3DApplication()
{
    // Override base class members
    m_strWindowTitle         = _T("Frogger");
    m_d3dEnumeration.AppUsesDepthBuffer = TRUE;
	m_dwCreationWidth		 = GetPrivateProfileInt("Frogger", "Width", cWidth, "media\\test.ini");
	m_dwCreationHeight		 = GetPrivateProfileInt("Frogger", "Height", cHeight, "media\\test.ini");

	m_Frogger = new CFrogger();
}


//-----------------------------------------------------------------------------
// Name: OneTimeSceneInit()
// Desc: Called during initial app startup, this function performs all the
//       permanent initialization.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::OneTimeSceneInit()
{
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FrameMove()
// Desc: Called once per frame, the call is the entry point for animating
//       the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FrameMove()
{
    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Called once per frame, the call is the entry point for 3d
//       rendering. This function sets up render states, clears the
//       viewport, and renders the scene.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::Render()
{
	m_pd3dDevice->SetRenderState(D3DRS_ZENABLE, D3DZB_FALSE);
	m_pd3dDevice->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
	m_pd3dDevice->SetRenderState(D3DRS_ALPHATESTENABLE, FALSE);
    // Begin the scene
    if (SUCCEEDED(m_pd3dDevice->BeginScene()))
    {
		// Clear the viewport
		m_pd3dDevice->Clear(0L, NULL, D3DCLEAR_TARGET|D3DCLEAR_ZBUFFER, 0x000000FF, 1.0f, 0L);

		m_Frogger->Render(GetTickCount());

		// End the scene.
        m_pd3dDevice->EndScene();
    }

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InitDeviceObjects()
// Desc: Initialize device-dependent objects. This is the place to create mesh
//       and texture objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InitDeviceObjects()
{
	// Create and load scene description
	if (FAILED(m_Frogger->Create(m_pd3dDevice, cWidth, cHeight)))
		return E_FAIL;

	m_Frogger->StartGame(GetTickCount()); // FIXME: shouldn't be here

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: RestoreDeviceObjects()
// Desc: Restore device-memory objects and state after a device is created or
//       resized.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::RestoreDeviceObjects()
{
	if (FAILED(m_Frogger->Restore(m_d3dsdBackBuffer.Width, m_d3dsdBackBuffer.Height)))
		return E_FAIL;

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: InvalidateDeviceObjects()
// Desc: Called when the device-dependent objects are about to be lost.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::InvalidateDeviceObjects()
{
	if (FAILED(m_Frogger->Invalidate()))
		return E_FAIL;

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: DeleteDeviceObjects()
// Desc: Called when the app is exiting, or the device is being changed,
//       this function deletes any device dependent objects.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::DeleteDeviceObjects()
{
	m_Frogger->Destroy();

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: FinalCleanup()
// Desc: Called before the app exits, this function gives the app the chance
//       to cleanup after itself.
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::FinalCleanup()
{
    SAFE_DELETE(m_Frogger);

    return S_OK;
}

//-----------------------------------------------------------------------------
// Name: ConfirmDevice()
// Desc: Called during device initialization, this code checks the device
//       for some minimum set of capabilities
//-----------------------------------------------------------------------------
HRESULT CMyD3DApplication::ConfirmDevice(D3DCAPS9* pCaps, DWORD dwBehavior,
										 D3DFORMAT adapterFormat, D3DFORMAT backBufferFormat)

{
	if (dwBehavior & D3DCREATE_PUREDEVICE)
		return E_FAIL;

	if (pCaps->NumSimultaneousRTs < RT_NEEDED)
		return E_FAIL;

    if (pCaps->PixelShaderVersion < D3DPS_VERSION(2, 0))
        return E_FAIL;

	return S_OK;
}

//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: Message proc function to handle key and menu input
//-----------------------------------------------------------------------------
LRESULT CMyD3DApplication::MsgProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_KEYDOWN:
		m_Frogger->KeyDown(wParam);
		return 0;
	case WM_CONTEXTMENU:
		// Trap context menu
		return 0;
	}

    return CD3DApplication::MsgProc(hWnd, uMsg, wParam, lParam);
}

//-----------------------------------------------------------------------------

CMyD3DApplication g_d3dApp;

//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: Entry point to the program. Initializes everything, and goes into a
//       message-processing loop. Idle time is used to render the scene.
//-----------------------------------------------------------------------------
INT WINAPI WinMain(HINSTANCE hInst, HINSTANCE, LPSTR, INT)
{
    InitCommonControls();
    if (FAILED(g_d3dApp.Create(hInst)))
        return 0;

    return g_d3dApp.Run();
}
