#include <D3D9.h>
#include <D3DX9.h>
#include "DXUtil.h"
#include "D3DEnumeration.h"
#include "D3DSettings.h"
#include "D3DApp.h"

#include "Frogger.h"

const DWORD TVertex::FVF = D3DFVF_XYZ | D3DFVF_TEX3;


#define CHECK_RETURN if (FAILED(hr)) return hr;


CFrogger::CFrogger()
: m_Device(NULL), m_Reset(false), m_Effect(NULL), m_Quad(NULL),
  m_Background(NULL), m_Traffic(NULL), m_TrafficSpeed(NULL), m_Frogger(NULL), m_BabyFrog(NULL), m_GameOver(NULL),
  m_MiniTraffic(NULL), m_DirFromKey(NULL), m_DefaultRenderSurface(NULL), m_SinCos(NULL),
  m_Width(0), m_Height(0), m_StartTime(0), m_ActiveSelect(0), m_Key(0)
{
	m_Pos[0] = m_Pos[1] = NULL;
	m_GameData[0] = m_GameData[1] = NULL;
	m_Babies[0] = m_Babies[1] = NULL;
	m_TargetSurface[0] = m_TargetSurface[1] = NULL;
}


void CFrogger::FillBabies()
{
	// I wanted to create this in an art program, but couldn't find a single one that could edit single pixel RGBA well.
	// It was one of the most frustrating experiences I've had lately. Even the DX Texture tool didn't work right.
	// Note: The texture is of size 16 because the docs say that textures that are not power of 2 can't be accessed
	//       with a dependent read.
	IDirect3DSurface9 * Surface, * DestSurface;
	m_Device->CreateOffscreenPlainSurface(16, 1, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &Surface, NULL);
	DWORD Colours[16] =
	{ 0, D3DCOLOR_XRGB(16,128,232), 0,0, D3DCOLOR_XRGB(232,160,16), 0,0, D3DCOLOR_XRGB(232,16,192),
	  0,0, D3DCOLOR_XRGB(16,232,16), 0,0, D3DCOLOR_XRGB(232,16,40), 0,0};
	D3DLOCKED_RECT TexData;
	Surface->LockRect(&TexData, NULL, 0);
	memcpy(TexData.pBits, Colours, 16*sizeof(DWORD));
	Surface->UnlockRect();
	m_Babies[0]->GetSurfaceLevel(0, &DestSurface);
	m_Device->StretchRect(Surface, NULL, DestSurface, NULL, D3DTEXF_NONE);
	DestSurface->Release();
	Surface->Release();
}

HRESULT CFrogger::Create(LPDIRECT3DDEVICE9 inDevice, DWORD inWidth, DWORD inHeight)
{
	int i;
	D3DLOCKED_RECT TexData;

	m_Device = inDevice;
	m_Width = inWidth;
	m_Height = inHeight;

	HRESULT hr;
	TVertex * Data;

    LPD3DXBUFFER Error = NULL;
    hr = D3DXCreateEffectFromFile(m_Device, "shaders\\frogger.fx", NULL, NULL, D3DXSHADER_DEBUG, NULL, &m_Effect, &Error);
	SAFE_RELEASE(Error);
	CHECK_RETURN(hr);

	hr = m_Effect->ValidateTechnique("Init");
	CHECK_RETURN(hr);
	hr = m_Effect->ValidateTechnique("Move");
	CHECK_RETURN(hr);
	hr = m_Effect->ValidateTechnique("Babies");
	CHECK_RETURN(hr);
	hr = m_Effect->ValidateTechnique("FinishUpdate");
	CHECK_RETURN(hr);
	hr = m_Effect->ValidateTechnique("Display");
	CHECK_RETURN(hr);

//	D3DXVECTOR4 ScreenSize;
//	m_Effect->GetVector("ScreenSize", &ScreenSize);

	if (FAILED(D3DXCreateTextureFromFile(m_Device, "media\\background.png", &m_Background)))
        return D3DAPPERR_MEDIANOTFOUND;

	if (FAILED(D3DXCreateVolumeTextureFromFile(m_Device, "media\\traffic.dds", &m_Traffic)))
        return D3DAPPERR_MEDIANOTFOUND;

	if (FAILED(D3DXCreateVolumeTextureFromFile(m_Device, "media\\mini-traffic.dds", &m_MiniTraffic)))
	// Screen size is specified, because otherwise D3DX rounds the width to 16
//	if (FAILED(D3DXCreateTextureFromFileEx(m_Device, "media\\mini-traffic.bmp", (int)ScreenSize[0], (int)ScreenSize[1], 1,
//		0, D3DFMT_UNKNOWN, D3DPOOL_MANAGED, D3DX_FILTER_NONE, D3DX_FILTER_NONE, 0, NULL, NULL, &m_MiniTraffic)))
        return D3DAPPERR_MEDIANOTFOUND;

	if (FAILED(D3DXCreateTextureFromFile(m_Device, "media\\traffic-speed.bmp", &m_TrafficSpeed)))
        return D3DAPPERR_MEDIANOTFOUND;

	if (FAILED(D3DXCreateTextureFromFile(m_Device, "media\\Frogger.png", &m_Frogger)))
        return D3DAPPERR_MEDIANOTFOUND;

	if (FAILED(D3DXCreateTextureFromFile(m_Device, "media\\Baby.png", &m_BabyFrog)))
        return D3DAPPERR_MEDIANOTFOUND;

	if (FAILED(D3DXCreateTextureFromFile(m_Device, "media\\GameOver.png", &m_GameOver)))
        return D3DAPPERR_MEDIANOTFOUND;

	// Position (and direction) (two textures, one serves for reading, the other as render target; switched each frame)
	m_Device->CreateTexture(1, 1, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, &m_Pos[0], NULL);
	m_Device->CreateTexture(1, 1, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, &m_Pos[1], NULL);
	// Extra game data (lives, ...)
	m_Device->CreateTexture(1, 1, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, &m_GameData[0], NULL);
	m_Device->CreateTexture(1, 1, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, &m_GameData[1], NULL);
		// D3DFMT_A32B32G32R32F for convenience of assignment (16 bit would be enough for accuracy)
	// FIXME: verify that the format exists (although can probably be assumed if PS2 exists)

	m_Device->CreateTexture(16, 1, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_Babies[0], NULL);
	m_Device->CreateTexture(16, 1, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_Babies[1], NULL);
	FillBabies();

	// Create the key to direction texture
	// For some reason, even though the format is ABGR, XYZW of the vector correspond to RGBA.
	// RG are (X,Y) delta, B is direction counterclockwise from 0 at top with 0.5 at bottom, A is 0.
	m_Device->CreateTexture(256, 1, 1, 0, D3DFMT_A32B32G32R32F, D3DPOOL_MANAGED, &m_DirFromKey, NULL);
	m_DirFromKey->LockRect(0, &TexData, NULL, 0);
	for (i=0; i<256; i++)
		((D3DXVECTOR4*)TexData.pBits)[i] = D3DXVECTOR4(0,0,0,0);
	((D3DXVECTOR4*)TexData.pBits)['W'] = ((D3DXVECTOR4*)TexData.pBits)[VK_UP] = D3DXVECTOR4(0,-1, 0, 0);
	((D3DXVECTOR4*)TexData.pBits)['S'] = ((D3DXVECTOR4*)TexData.pBits)[VK_DOWN] = D3DXVECTOR4(0,1, 0.5, 0);
	((D3DXVECTOR4*)TexData.pBits)['A'] = ((D3DXVECTOR4*)TexData.pBits)[VK_LEFT] = D3DXVECTOR4(-1,0, 0.25, 0);
	((D3DXVECTOR4*)TexData.pBits)['D'] = ((D3DXVECTOR4*)TexData.pBits)[VK_RIGHT] = D3DXVECTOR4(1,0, 0.75, 0);
	m_DirFromKey->UnlockRect(0);


	// Create the sin-cos texture.
	// The texture contains (cos, sin, -sin), so that it can be easily used for rotation.
	// It's low accuracy -- only 64 values -- but I don't really use them much anyway.
	m_Device->CreateTexture(64, 1, 1, 0, D3DFMT_A32B32G32R32F, D3DPOOL_MANAGED, &m_SinCos, NULL);
	m_SinCos->LockRect(0, &TexData, NULL, 0);
	for (i=0; i<64; i++)
	{
		float Angle = i / 64.0f * 2 * D3DX_PI;
		float Cos = (float)cos(Angle);
		float Sin = (float)sin(Angle);
		((D3DXVECTOR4*)TexData.pBits)[i] = D3DXVECTOR4(Cos,-Sin,0.5f,Sin);
	}
	m_SinCos->UnlockRect(0);



	hr = m_Effect->SetTexture("tBackground", m_Background);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tTraffic", m_Traffic);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tMiniTraffic", m_MiniTraffic);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tTrafficSpeed", m_TrafficSpeed);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tFrogger", m_Frogger);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tBabyFrog", m_BabyFrog);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tGameOver", m_GameOver);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tDirFromKey", m_DirFromKey);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tSinCos", m_SinCos);
	CHECK_RETURN(hr);


	hr = m_Device->CreateVertexBuffer(4*sizeof(TVertex), D3DUSAGE_WRITEONLY,
		TVertex::FVF, D3DPOOL_MANAGED, &m_Quad, NULL);
	CHECK_RETURN(hr);
	hr = m_Quad->Lock(0,0, (void**)&Data,0);
	CHECK_RETURN(hr);
	Data[0] = TVertex(-1,-1, 0,0, 0,0, 0,0);
	Data[1] = TVertex(-1,1, 0,0.75f, 0,0.75f, 0,12);
	Data[2] = TVertex(1,-1, 1,0, 0.5f,0, 16,0);
	Data[3] = TVertex(1,1, 1,0.75f, 0.5f,0.75f, 16,12);
	hr = m_Quad->Unlock();

	// FIXME: 16 and 12 above should be a constant and set in the shader


	hr = m_Device->GetRenderTarget(0, &m_DefaultRenderSurface);

	return S_OK;
}


HRESULT CFrogger::Destroy()
{
    SAFE_RELEASE(m_Quad);
	SAFE_RELEASE(m_Effect);
    SAFE_RELEASE(m_Background);
    SAFE_RELEASE(m_Traffic);
    SAFE_RELEASE(m_MiniTraffic);
    SAFE_RELEASE(m_TrafficSpeed);
    SAFE_RELEASE(m_Frogger);
    SAFE_RELEASE(m_BabyFrog);
    SAFE_RELEASE(m_GameOver);
    SAFE_RELEASE(m_DirFromKey);
    SAFE_RELEASE(m_SinCos);
    SAFE_RELEASE(m_Pos[0]);
    SAFE_RELEASE(m_Pos[1]);
    SAFE_RELEASE(m_GameData[0]);
    SAFE_RELEASE(m_GameData[1]);
    SAFE_RELEASE(m_Babies[0]);
    SAFE_RELEASE(m_Babies[1]);
	SAFE_RELEASE(m_DefaultRenderSurface);

	return S_OK;
}


HRESULT CFrogger::Invalidate()
{
	// The game is reset if the device needs to be reset. That's because all the game data
	// resides in on-card textures.
	for (int i=0; i<2; i++)
	{
		SAFE_RELEASE(m_Pos[i]);
		SAFE_RELEASE(m_GameData[i]);
		SAFE_RELEASE(m_Babies[i]);
	}
	SAFE_RELEASE(m_DefaultRenderSurface);
	m_Effect->OnLostDevice();
	m_Reset = true;
	return S_OK;
}

HRESULT CFrogger::Restore(DWORD width, DWORD height)
{
	// The framework has a strange way of handling resets, by releasing and recreating objects.
	// I'd rather do correct reset handling, so I mark it.
	if (m_Reset)
	{
		for (int i=0; i<2; i++)
		{
			m_Device->CreateTexture(1, 1, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, &m_Pos[i], NULL);
			m_Device->CreateTexture(1, 1, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A32B32G32R32F, D3DPOOL_DEFAULT, &m_GameData[i], NULL);
			m_Device->CreateTexture(16, 1, 1, D3DUSAGE_RENDERTARGET, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT, &m_Babies[i], NULL);
		}
		FillBabies();
		m_Device->GetRenderTarget(0, &m_DefaultRenderSurface);
		m_Effect->OnResetDevice();
		m_ActiveSelect = 0;
		m_NewGame = true;
	}
	return S_OK;
}


void CFrogger::SetMoveRenderTarget(int inTextureSelect)
{
    m_Pos[inTextureSelect]->GetSurfaceLevel(0, &m_TargetSurface[0]);
    m_GameData[inTextureSelect]->GetSurfaceLevel(0, &m_TargetSurface[1]);
	for (int i=0; i<RT_NEEDED; i++)
		m_Device->SetRenderTarget(i, m_TargetSurface[i]);
}

void CFrogger::SetBabiesRenderTarget(int inTextureSelect)
{
    m_Babies[inTextureSelect]->GetSurfaceLevel(0, &m_TargetSurface[0]);
	m_Device->SetRenderTarget(0, m_TargetSurface[0]);
}


void CFrogger::RestoreRenderTarget() // restore rendering to back buffer
{
	m_Device->SetRenderTarget(0, m_DefaultRenderSurface);
	for (int i=0; i<RT_NEEDED; i++)
	{
		if (i > 0)
			m_Device->SetRenderTarget(i, NULL);
		if (m_TargetSurface[i] != NULL)
		{
			m_TargetSurface[i]->Release();
			m_TargetSurface[i] = NULL;
		}
	}
}



void CFrogger::StartGame(DWORD inTime)
{
	m_PrevTime = m_StartTime = inTime;
	m_ActiveSelect = 0;
	m_NewGame = true;
}

void CFrogger::KeyDown(WPARAM inKey)
{
	m_Key = inKey;
}


HRESULT CFrogger::RenderTechnique()
{
	UINT iPass, Passes;
	m_Effect->Begin(&Passes, 0);
	for (iPass = 0; iPass < Passes; iPass++)
	{
		m_Effect->Pass(iPass);
		m_Device->DrawPrimitive(D3DPT_TRIANGLESTRIP, 0, 2);
	}
	m_Effect->End();
	return S_OK;
}

HRESULT CFrogger::Render(DWORD inTime)
{
    HRESULT hr;

	m_Device->SetVertexDeclaration(NULL);
	m_Device->SetFVF(TVertex::FVF);
	m_Device->SetStreamSource(0, m_Quad, 0, sizeof(TVertex));


	if (m_NewGame)
	{
		SetMoveRenderTarget(0);
		hr = m_Effect->SetTechnique("Init");
		CHECK_RETURN(hr);
		RenderTechnique();
		RestoreRenderTarget();
		m_NewGame = false;
	}

	hr = m_Effect->SetFloat("TimeFromPrevFrame", (inTime - m_PrevTime) / 1000.0f);

	D3DXVECTOR4 Key(m_Key / 256.0f, 0, 0, 0);
	m_Effect->SetVector("KeyPress", &Key);
//	m_Effect->SetFloat("KeyPress", m_Key / 256.0f);

	hr = m_Effect->SetTechnique("Move");
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tPos", m_Pos[0]);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tGameData", m_GameData[0]);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tBabies", m_Babies[m_ActiveSelect]);
	CHECK_RETURN(hr);
	SetMoveRenderTarget(1);
	RenderTechnique();
	RestoreRenderTarget();

	hr = m_Effect->SetTechnique("Babies");
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tPos", m_Pos[1]);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tGameData", m_GameData[1]);
	CHECK_RETURN(hr);
	SetBabiesRenderTarget(1-m_ActiveSelect);
	RenderTechnique();
	RestoreRenderTarget();

	hr = m_Effect->SetTechnique("FinishUpdate");
	CHECK_RETURN(hr);
	SetMoveRenderTarget(0);
	RenderTechnique();
	RestoreRenderTarget();

	hr = m_Effect->SetTechnique("Display");
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tPos", m_Pos[0]);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tGameData", m_GameData[0]);
	CHECK_RETURN(hr);
	hr = m_Effect->SetTexture("tBabies", m_Babies[1-m_ActiveSelect]);
	CHECK_RETURN(hr);
	RenderTechnique();

	m_PrevTime = inTime;
	m_ActiveSelect = 1 - m_ActiveSelect;
	m_Key = 0;

	return S_OK;
}
