// g:fxc /T ps_2_0 /E Display /Fc display.ps frogger.fx

texture tBackground < string name = "..\media\background.png"; >;
texture tTraffic;
texture tMiniTraffic;
texture tTrafficSpeed;
texture tFrogger;
texture tBabyFrog;
texture tGameOver;
texture tDirFromKey;
texture tSinCos;
texture tPos;
texture tGameData;
texture tBabies;

// All samplers use point sampling for now.
// The only ones that really need point sampling are the 2D screen resolution ones,
// such as sTrafficSpeed.

sampler sBackground = sampler_state
{
	Texture = (tBackground);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Wrap; // Doesn't matter -- will always be 0 to 1
	AddressV = Wrap; // Doesn't matter
};

sampler sTraffic = sampler_state
{
	Texture = (tTraffic);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Wrap;
	AddressV = Clamp; // Doesn't matter -- will always be 0 to 1
	AddressW = Mirror;
};

sampler sMiniTraffic = sampler_state
{
	Texture = (tMiniTraffic);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Wrap;
	AddressV = Clamp; // Doesn't matter -- will always be 0 to 1
	AddressW = Mirror;
};

sampler sTrafficSpeed = sampler_state
{
	Texture = (tTrafficSpeed);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Wrap; // Doesn't matter -- will always be 0 to 1
	AddressV = Wrap; // Doesn't matter
};

sampler sFrogger = sampler_state
{
	Texture = (tFrogger);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Clamp; // So that frogger shows only at its position.
	AddressV = Clamp; // (Note that frogger's edge should be with alpha=0)
};

sampler sBabyFrog = sampler_state
{
	Texture = (tBabyFrog);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Wrap; // Baby frogs show along and entire line
	AddressV = Clamp;
};

sampler sGameOver = sampler_state
{
	Texture = (tGameOver);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Clamp;
	AddressV = Clamp;
};

sampler sDirFromKey = sampler_state
{
	Texture = (tDirFromKey);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Clamp; // Key is a WORD (WPARAM), so might be higher than 1
	AddressV = Clamp; // Doesn't matter
};

sampler sSinCos = sampler_state
{
	Texture = (tSinCos);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Wrap;
	AddressV = Wrap; // Doesn't matter
};

sampler sPos = sampler_state // position texture
{
	Texture = (tPos);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Clamp; // Doesn't matter -- will always be 0 to 1
	AddressV = Clamp; // Doesn't matter
};

sampler sGameData = sampler_state
{
	Texture = (tGameData);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Clamp; // Doesn't matter -- will always be 0 to 1
	AddressV = Clamp; // Doesn't matter
};

sampler sBabies = sampler_state
{
	Texture = (tBabies);
	MinFilter = Point;
	MagFilter = Point;
	AddressU = Clamp;
	AddressV = Clamp; // Doesn't matter (height is one texel)
};



static const float4 PreStartGameData = float4(5.5f, -1, 0, 1.0f/4); // 5 lives, game over, 5 baby frogs, "Frogger" message
static const float4 InitialGameData = float4(5.5f, 100, 0, 0); // 5 lives and 5 baby frogs, 100 seconds game time, scroll pos 0, no message
// Note: the following all depend on the game screen size (16x12), which is dependent
//       on the art and the C++ code, and should not be changed.
static const float2 ScreenSize = float2(16,12);
static const float4 InitialPosition = float4(0.5f,11.5f/16, 0, 0); // Frogger centre at (8,11.5), facing up, not death scene
//static const float4 InitialPosition = float4(0.5f,11.5f/16, 0, 0); // Frogger centre at (8,11.5), facing up, not death scene
static const float2 TopLeftBorder = float2(1.0f/16, 0.5f/16);
static const float2 BottomRightBorder = float2(15.0f/16, 11.5f/16);
static const float2 FrogStep = (1/16.0f, 1/16.0f);

float2 KeyPress; // warning of dependent read if just 'float'
float TimeFromPrevFrame;


struct Input
{
	float2 ScreenPos: TEXCOORD0;
	float2 TrafficPos: TEXCOORD1;
	float2 GamePos: TEXCOORD2;
};

struct OutputMove
{
	float4 Pos: COLOR0;
	float4 Data: COLOR1;
};




OutputMove Init(Input In) : COLOR
{
	OutputMove Out;
	Out.Pos = InitialPosition;
	Out.Data = PreStartGameData;
	return Out;
}


// The data is kept in two 1x1 samplers: sPos and sGameData.
// sPos.xy is frogger's game position; sPos.z is his facing direction; sPos.w is used for the death rotation
// sGameData.x = lives
// sGameData.y = remaining time
// [Frogs left - float)]
// Note: all comparisons use inequalities even when equalities are clearer, to save instructions.
OutputMove Move(Input In)
{
	OutputMove Out;

	float4 Pos = tex1D(sPos, In.ScreenPos); // sampling position doesn't matter
	float4 Data = tex1D(sGameData, In.ScreenPos); // sampling position doesn't matter

	// Check for running out of time (TimeUp will be used later).
	float TimeUp = Data.y * (Data.y - TimeFromPrevFrame);
		// If the time moved from positive to negative, or to/from 0, TimeUp will be negative,
		// to mark Game Over. Other reasons for ending the game set the time to negative, so
		// the check will not set "Game Over" again.
	Data.y -= TimeFromPrevFrame;
	Data.z += TimeFromPrevFrame * (1 - frac(Data.x)); // "1-frac(Data.x)" is speed, growing as babies are rescued

	float4 NewPos = Pos;
	float4 NewData = Data;

	if (Data.y < 0)
	{
		if (KeyPress.x * (-0.3f - Data.y) > 0)
//		if (KeyPress.x > 0 && Data.y < -1) // Game Over status (wait one second after game over)
			NewPos.w = -4; // reset position and game data
	}
	else
	{
		// Calculate new position based on key press
		float4 Dir4 = tex1D(sDirFromKey, KeyPress); // dep1
		NewPos.xy = min(max(NewPos.xy + Dir4.xy * FrogStep, TopLeftBorder), BottomRightBorder); // Don't exit the screen
		// Note: Compiler wouldn't generate both 'min' and 'max' when babies code wasn't after this.
		if (dot(Dir4, Dir4) > 0) // Dir4 is all 0 only for no move
			NewPos.z = Dir4.z;



		// Sample babies texture. This is done using the original position (row doesn't matter),
		// and then tested using the Y of the new position to see if the test is applicable.
		// This is to reduce dependency.
		float4 TopRow = tex2D(sBabies, Pos - float2(0.5f/16.0f, 1/16.0f)); // dep1
		float PosY = NewPos.y - 1/16.0f;
		// If the baby frog is there (alpha
		// TopRow.a is either 0 or 1, so  -PosY * TopRow.a > 0  iff  -PosY > 0 && TopRow.y > 0
		if (PosY * TopRow.a < 0) // If the baby frog is there (alpha
//		if (NewPos.y < 1/16.0f && TopRow.a > 0)
		{
			NewData.x = Data.x - 0.1f; // reduce number of baby frogs left
			NewPos.w = -3; // set initial position
		}
		if (PosY + dot(TopRow.xyz, TopRow.xyz) <= 0)
//		if (NewPos.y <= 1/16.0f && dot(TopRow.xyz, TopRow.xyz) <= 0) // if on a barrier
			NewPos.xy = Pos.xy; // move back to where you were



		// Snap frogger to the grid.
		// This scrolls frogger in scrolling rows (water), and snaps him to the right place
		// when moving between rows.
		float2 Scroll = tex2D(sTrafficSpeed, NewPos).rg; // dep2, saves 1 instruction over the next
	//	float2 Scroll = tex1D(sTrafficSpeed, NewPos.y).rg;
		float ScrollXTime = (Scroll.x - 0.5f) * Data.z;
			// Scroll.y is either 0 (no scroll) or 1 (scroll)
		float Frac = frac(NewPos.x * ScreenSize.x - (ScrollXTime * ScreenSize.x * 2 - 0.5f) * Scroll.y);
			// Scroll.y is either 0 (no scroll) or 1 (scroll)
		if (Frac > 0.5f)
			Frac -= 1;
		NewPos.x -= Frac / ScreenSize.x;



		// Test for hit.
		// Get the pixel from the mini game screen (scrolled according to the
		// time). If this pixel isn't 0, frogger has hit something, and his
		// position and direction are reset.
		// (Note that in the water, the logs have zero and the water has 1.)
		float3 CheckPos;
		CheckPos.x = NewPos.x / 2 - ScrollXTime;// + 0.5f/16;
			// The '-' is used because the positive scroll direction is right,
			// which means that a previous texel (one to the left) must be used.
		CheckPos.y = NewPos.y - 0.5f/16;
		CheckPos.z = Data.z / 2;
		float Hit = tex3D(sMiniTraffic, CheckPos).r; // dep3
		NewData.x -= Hit; // Lives. Hit will be 0 (no hit) or 1 (hit).
		if (Hit > 0) // was == 1 (but inequalities and comparisons to 0 use fewer instructions)
			NewPos.w = 2; // mark as death scene
	}

	if (TimeUp <= 0)
		NewData.w = 3.0f/4; // "Game Over"

	if (Pos.w > 0) // is > 0 during the death animation, 0 normally
	{
		NewPos = Pos;
		NewData = Data;
		NewPos.w -= TimeFromPrevFrame * 2;
		if (NewPos.w < 0)
			NewPos.w = -3; // set initial position
	}

	Out.Pos = NewPos;
	Out.Data = NewData;

	return Out;
}


// Update the frog babies texture.
// The babies' colours are always kept in the texture, and their alpha tells if they're shown.
// Empty slots are completely black (0,0,0,0).
float4 Babies(Input In) : COLOR
{
	In.ScreenPos += float2(0.5f, 0.5f) / ScreenSize; // move to centre of texel for correct fetching!
	float4 Out = tex1D(sBabies, In.ScreenPos.x);

	float4 Pos = tex2D(sPos, In.ScreenPos); // sampling position doesn't matter
	float4 Data = tex2D(sGameData, In.ScreenPos); // sampling position doesn't matter

	// If frogger is on this position, make the frog invisible
	if (Pos.y <= 1/16.0f && floor(In.GamePos.x) + 1 == floor(Pos.x * ScreenSize.x))
		// Change <= 0 to <=0.5 to make sure no accuracy problem?
		Out.a = 0; // could use fewer components per texture (but doesn't really matter)

	// If number of babies equals the initial number, show all frogs
	// (The Move shader updates the count before this shader is executed, so if
	//  Frogger is on a baby, the count will always be less than the initial.)
	if (frac(Data.x) >= frac(InitialGameData.x) && dot(Out,Out) > 0)
		Out.a = 1;

	return Out;
}


OutputMove Move2(Input In)
{
	OutputMove Out;

	float4 Pos = tex1D(sPos, In.ScreenPos); // sampling position doesn't matter
	float4 Data = tex1D(sGameData, In.ScreenPos); // sampling position doesn't matter

	// If no babies left
	if (Data.y >= 0)
	{
		if (frac(Data.x) <= 0.01f) // due to lack of accuracy (I guess), comparison to 0 doesn't work
		{
			Data.w = 2.0f/4; // "You Win!"
			Data.y = -0.001f; // end game
		}
		
		// If no lives left (but don't stop death animation)
		if (max(floor(Data.x), -Pos.w) <= 0) // saves 3 instructions over the next
//		if (floor(Data.x) <= 0 && Pos.w >= 0)
		{
			Data.w = 3.0f/4; // "Game Over"
			Data.y = -0.001f; // end game
		}
	}

	if (Pos.w <= -3)
		Out.Pos = InitialPosition;
	else
		Out.Pos = Pos;

	if (Pos.w <= -4)
	{
		float RandomPos = frac(Data.y) * 10; // for a random scroll start
		Out.Data = InitialGameData;
		Out.Data.z = RandomPos;
	}
	else
		Out.Data = Data;
		
	return Out;
}



// Render the background (scrolled) and frogger over it, facing according to FrogMove.
// ScreenPos runs from 0 to 1 in both directions
float4 Display(Input In) : COLOR
{
//	return tex2D(sBabies, ScreenPos) * tex2D(sBabyFrog, ScreenPos * ScreenSize);

	float4 FrogPos = tex1D(sPos, In.ScreenPos); // sampling position doesn't matter
	float4 SinCos = tex1D(sSinCos, FrogPos.z + FrogPos.w);
	float4 Data = tex1D(sGameData, In.ScreenPos); // sampling position doesn't matter

	float4 Colour = tex2D(sBackground, In.ScreenPos);
	float4 Colour2;

	// Note: Time is in seconds, full scroll is 1, so (sTrafficSpeed-0.5) encodes full scrolling per second.
	//       For example, if a texel is 0.2+0.5 (~179/255), then the lane will fully scroll right every five
	//       seconds, and if it's -0.1+0.5 (102/255), the the lane would fully scroll left every ten seconds.
	float3 ScrollPos;
	ScrollPos.x = In.TrafficPos.x - (tex2D(sTrafficSpeed, In.TrafficPos).r - 0.5f) * Data.z; // saves over the next
//	ScrollPos.x = In.TrafficPos.x - (tex1D(sTrafficSpeed, In.TrafficPos.y).r - 0.5f) * Data.z;
		// The '-' is used because the positive scroll direction is right,
		// which means that a previous texel (one to the left) must be used.
	ScrollPos.y = In.TrafficPos.y;
	ScrollPos.z = Data.z / 2;
	Colour2 = tex3D(sTraffic, ScrollPos);
	Colour = lerp(Colour, Colour2, Colour2.a);

	// Draw the baby frogs at the top
	// OPT: a set of texture coords can have the value already subtracted
	Colour2 = tex1D(sBabies, In.ScreenPos - float2(0.5f/16,0)) * tex2D(sBabyFrog, (In.GamePos - float2(0.5f,0)));
	Colour = lerp(Colour, Colour2, Colour2.a);


	// Draw a time bar at the bottom right
	float2 BarPos = In.ScreenPos - float2(0.65f, 0.73f);
	if (min(Data.y * 0.003f - BarPos.x, min(BarPos.y, BarPos.x)) >= 0)
		// Changing the order of arguments within a 'min' can change the shader completely.
		// Here 'Data.y * 0.003f - BarPos.x' was moved to the start because it generates
		// a shader where 'texld' instructions are less contiguous, which should work
		// faster on ATI hardware.
//	if (In.ScreenPos.y > 0.73f && PosX >= 0 && PosX < Data.y * 0.003f)
		Colour = float4(1,0,1,1); // purple

	// Flip frogger's texture coordinate over its middle according to FrogMove
	// Frogger faces up when x==0 and y==-1, right when x==1 and y==0
	float2 DistFromFrogMiddle = In.GamePos - FrogPos.xy / FrogStep;
	// SinCos contains (cos, -sin, 0.5f, sin)
	// [0.5f is there for 'dot', but HLSL compiler doesn't like swizzling --
	//  even when it can generate the same swizzling in other cases.
	//  Note: this makes it less optimal for ATI hardware, since scalar and
	//        vector parts end up being used at the same time.
	float2 FrogFacing = float2(0.5f,0.5f) + DistFromFrogMiddle.x * SinCos.xw + DistFromFrogMiddle.y * SinCos.yx;
		// Note: need the constant at the start for optimal 'mad' sequence to be generated
	Colour2 = tex2D(sFrogger, FrogFacing);
	Colour = lerp(Colour, Colour2, Colour2.a);

	// Draw number of lives (top left of screen)
	if (In.ScreenPos.x / 0.03f < floor(Data.x)) // faster because matches a later calculation
//	if (In.ScreenPos.x < floor(Data.x) * 0.03f)
	{
		float2 SamplePos;
		SamplePos.y = In.ScreenPos.y / 0.03f;
		SamplePos.x = frac(In.ScreenPos.x / 0.03f);
		Colour2 = tex2D(sFrogger, SamplePos);
		Colour = lerp(Colour, Colour2, Colour2.a);
	}

	// Draw blinking message (Game Over, etc.). Data.w selects the message.
	if (frac(Data.y) >= 0.5f)
	{
		float2 MessagePos = { In.ScreenPos.x, Data.w + saturate(In.ScreenPos.y*4 - 1) * 0.25 };
		Colour2 = tex2D(sGameOver, MessagePos);
		Colour = lerp(Colour, Colour2, Colour2.a);
	}

	return Colour;
}



technique Init
{
    pass P0
    {        
        PixelShader  = compile ps_2_0 Init();
        fvf          = XYZ | Tex3;
        vertexshader = NULL;
 
        Lighting     = False;
        CullMode     = None;
    }
}

technique Move
{
    pass P0
    {        
        PixelShader  = compile ps_2_0 Move();
        fvf          = XYZ | Tex3;
        vertexshader = NULL;
 
        Lighting     = False;
        CullMode     = None;
    }
}

technique Babies
{
    pass P0
    {        
        PixelShader  = compile ps_2_0 Babies();
        fvf          = XYZ | Tex3;
        vertexshader = NULL;
 
        Lighting     = False;
        CullMode     = None;
    }
}

technique FinishUpdate
{
    pass P0
    {        
        PixelShader  = compile ps_2_0 Move2();
        fvf          = XYZ | Tex3;
        vertexshader = NULL;
 
        Lighting     = False;
        CullMode     = None;
    }
}

technique Display
{
    pass P0
    {        
        PixelShader  = compile ps_2_0 Display();
        fvf          = XYZ | Tex3;
        vertexshader = NULL;
 
        Lighting     = False;
        CullMode     = None;
    }
}
